#include <iostream>
#include <cstdint>
#include <map>
#include <vector>
#include <string>
#include <sstream>
#include <ctype.h>

using std::map;
using std::string;
using std::vector;

const int MBMEM = 32;

map<string, int> labels;
map<string, int> dep_labels;
enum token_type {
	LINENUM,
	LABEL_REF,
	LABEL_DEF,
	PSEUDO,
	JMP_PSEUDO,
	INSTR,
	NUMBER
};
struct token_t {
	token_type type;
	string s_value;
	int i_value;
};
vector<vector<token_t>> source;
vector<token_t>  work[MBMEM];
uint32_t out[MBMEM];

std::ostream& operator << (std::ostream&o, enum token_type t)
{
	switch(t) {
	case LINENUM:
		o << "LINENUM";
		break;
	case LABEL_REF:
		o << "LABEL_REF";
		break;
	case LABEL_DEF:
		o << "LABEL_DEF";
		break;
	case PSEUDO:
		o << "PSEUDO";
		break;
	case INSTR:
		o << "INSTR";
		break;
	case NUMBER:
		o << "NUMBER";
		break;
	case JMP_PSEUDO:
		o << "JMP_PSEUDO";
		break;
	}
	return o;
}

void dump()
{
	std::cout << "Source: \n";
	for (auto l: source) {
		for (auto t: l) {
			std::cout << "(" << t.type << ", " << t.s_value << ", " << t.i_value << ") ";
		}
		std::cout << "\n";
	}
}

void dump_labels()
{
	std::cout << "Labels: \n";
	for (auto l: labels) {
		std::cout << l.first << " = " << l.second << "\n";
	}
	std::cout << "Dependent labels: \n";
	for (auto l: dep_labels) {
		std::cout << l.first << " = " << l.second << "\n";
	}
}

void syntaxerror(int ln,  const string& message)
{
	string e = std::to_string(ln) + ": " + message;
	throw e;
}

bool islabel(const string& word)
{
	if (word.length() < 1)
		return false;
	if (!isalpha(word[0]) && word[0] != '_')
		return false;
	for (int i = 1; i < word.length(); i++)
		if (!isalpha(word[i]) && word[i] != '_' && word[i] != '@' && !isdigit(word[i]))
			return false;
	return true;
}

bool islabel_def(const string& word, token_t& token)
{
	size_t len = word.length();
	if (len < 2)
		return false;
	if (word.back() != ':')
		return false;
	string label = word.substr(0, len - 1);
	if (!islabel(label))
		return false;

	token.type = LABEL_DEF;
	token.s_value = label;
	token.i_value = 0;
	return true;
}

bool islabel_ref(const string& word, token_t& token)
{
	if (!islabel(word))
		return false;
	token.type = LABEL_REF;
	token.s_value = word;
	token.i_value = 0;
	return true;
}

bool isinstr(const string& word, token_t& token)
{
	int v = -1;
	if (word == "JMP") {
		v = 0;
	} else if (word == "JRP") {
		v = 1;
	} else if (word == "LDN") {
		v = 2;
	} else if (word == "STO") {
		v = 3;
	} else if (word == "SUB") {
		v = 4;
	} else if (word == "CMP") {
		v = 6;
	} else if (word == "STP") {
		v = 7;
	}
	if (v == -1)
		return false;
	token.type = INSTR;
	token.s_value = word;
	token.i_value = v;
	return true;
}

bool ispseudo(const string& word, token_t& token)
{
	if (word == ".org" || word == ".int") {
		token.type = PSEUDO;
		token.s_value = word;
		token.i_value = 0;
		return true;
	}
	return false;
}

bool isnumber(const string& word, token_t& token)
{
	int v = 0;
	try {
		v = stoi(word, nullptr);
	} catch (...) {
		return false;
	}
	token.type = NUMBER;
	token.s_value = word;
	token.i_value = v;
	return true;
}

int getaddr(const vector<token_t>& line, int tn)
{
	if (line.size() <= tn) {
		syntaxerror(line[0].i_value, "missing address");
	}
	if (line[tn].type != NUMBER) {
		syntaxerror(line[0].i_value, ".org must be followed by a number");
	}
	if (line[tn].i_value < 0 || line[tn].i_value >= MBMEM) {
		syntaxerror(line[0].i_value, "address " + std::to_string(line[tn].i_value) + " out of bounds");
	}
	return line[tn].i_value;
}

int getvalue(const vector<token_t>& line, int tn)
{
	if (line.size() <= tn) {
		return 0;
	}
	const token_t& t = line[tn];
	switch (t.type) {
	case NUMBER:
		return t.i_value;
	case LABEL_REF: {
			auto it = labels.find(t.s_value);
			if (it == labels.end())
				syntaxerror(line[0].i_value, "label '" + t.s_value + "' undefined");
			return it->second;
		}
	default:
		throw "Internal error";
	}		
	return 0;
}

const char *
bin2ascii(int32_t v)
{
	static char buf[33];
	uint32_t v_ = (uint32_t)v;

	for (int i = 0; i < 32; i++) {
		buf[i] = '0' + (v_ & 0x1);
		v_ >>= 0x1;
	}
	buf[32] = 0;
	return buf;
}

int main()
{
	/* first pass: tokenize and check syntax */
	try {
		string last_label;
		for (int ln = 1; !std::cin.eof(); ln++) {
			string line;
			getline(std::cin, line);
			std::istringstream iss(line);
			vector<token_t> tokens{token_t{LINENUM, "", ln}};
			enum { START, LABEL, OPT, END } state = START;
			do {
				string word;
				iss >> word;
				if (word[0] == ';')
					break;
				if (!word.length())
					continue;
				token_t token;
				if (state == LABEL) {
					source.push_back(tokens);
					tokens = vector<token_t>{token_t{LINENUM, "", ln}};
				}
				switch (state) {
				case START:
				case LABEL:
					if (islabel_def(word, token)) {
						last_label = word.substr(0, word.length() - 1);;
						state = LABEL;
					} else if (isinstr(word, token)) {
						state = OPT;
					} else if (word == ".jmp") {
						size_t end;
						if ((end = last_label.find_first_of('@')) == string::npos)
							syntaxerror(work[ln][0].i_value, "illegal use of .jmp");
						token.type = JMP_PSEUDO;
						token.s_value = last_label;
						token.i_value = end;
						state = END;
					} else if (ispseudo(word, token)) {
						state = OPT;
					} else {
						syntaxerror(ln, "unexpected '" + word + "'");
					}
					break;
				case OPT:
					if (isnumber(word, token))
						state = END;
					else if (islabel_ref(word, token))
						state = END;
					else
						syntaxerror(ln, "got '" + word + "' while expecting a number or a label");
					break;
				case END:
					syntaxerror(ln, "extra tokens starting at '" + word + "'");
					break;
				default:
					throw string("Internal error 1st pass");
				}
				tokens.push_back(token);
			} while (iss);
			if (tokens.size() > 1) {
				source.push_back(tokens);
			}
		}
	} catch (string e) {
		std::cerr << e << "\n";
		exit(1);
	}
	//dump();
	/* second pass: assign tokens to memory locations and define labels */
	try {
		int ln = 1;
		for (auto line: source) {
			const token_t& t = line[1];
			switch (t.type) {
			case PSEUDO:
				if (t.s_value == ".org") {
					ln = getaddr(line, 2);
				} else {
					work[ln] = line;
					ln++;
				}
				break;
			case JMP_PSEUDO:
				work[ln] = line;
				ln++;
				break;
			case INSTR:
				/* define dependent labels */
				if (line.size() > 2) {
					const token_t& l = line[2];
					if (l.type == LABEL_REF && l.s_value.find_first_of('@') != string::npos) {
						int v = (t.s_value == "JRP" ? ln : 0);
						auto it = dep_labels.find(l.s_value);
						if (it != dep_labels.end() && it->second != v) {
							syntaxerror(line[0].i_value, "Label '" + l.s_value + "' already used in an incompatible way");
						}
						dep_labels[l.s_value] = v;
					}
				}
				work[ln] = line;
				ln++;
				break;
			case LABEL_DEF:
				if (labels.find(t.s_value) != labels.end()) {
					syntaxerror(line[0].i_value, "label '" + t.s_value + "' already defined");
				}
				labels[t.s_value] = ln;
				break;
			default:
				throw string("Internal error 2nd pass");
			}
		}
	} catch (string e) {
		std::cerr << e << "\n";
		exit(1);
	}
	//dump_labels();
	/* third pass: assemble */
	try {
		for (int ln = 0; ln < MBMEM; ln++) {
			if (!work[ln].size()) {
				out[ln] = 0;
				continue;
			}
			const token_t& t = work[ln][1];
			switch (t.type) {
			case JMP_PSEUDO: {
				auto dit = dep_labels.find(t.s_value);
				if (dit == dep_labels.end())
					syntaxerror(work[ln][0].i_value, "dependend label '" + t.s_value + "' not referenced in a jmp instruction");
				string pl = t.s_value.substr(0, t.i_value);
				auto pit = labels.find(pl);
				if (pit == labels.end())
					syntaxerror(work[ln][0].i_value, "dependend label '" + t.s_value + "': parent label '" + pl + "' undefined");
				out[ln] = pit->second - dit->second - 1;
				break;
			}
			case PSEUDO: 
				out[ln] = getvalue(work[ln], 2);
				break;
			case INSTR:
				switch (t.i_value) {
				case 0:
				case 1:
				case 2:
				case 3:
				case 4:
				case 5:
					out[ln] = getvalue(work[ln], 2) | (t.i_value << 13);
					break;
				default:
					if (work[ln].size() > 2)
						syntaxerror(work[ln][0].i_value, "unexptected operand: '" + work[ln][2].s_value + "'");
					out[ln] = (t.i_value << 13);
					break;
				}
				break;
			default:
				throw string("Internal error 3rd pass");
			}
		}
	} catch (string e) {
		std::cerr << e << "\n";
		exit(1);
	}
	for (int ln = 0; ln < MBMEM; ln++)
		std::cout << bin2ascii(out[ln]) << "\n";
	return 0;
}
