#include <iostream>
#include <sstream>
#include <iomanip>
#include <chrono>
#include <cstdint>
#include <unistd.h>
#include <signal.h>
#include <sys/mman.h>
#include <cassert>
#include <capstone/capstone.h>

using std::chrono::time_point;
using std::chrono::steady_clock;
using std::chrono::duration_cast;
using std::chrono::microseconds;
using namespace std::literals::chrono_literals;

/* some global options */

/* verbosity level. Affects the detail of the output */
int verbose = 0;
/* perform a sleep after the execution of each instruction,
 * to emulate the running time on the original machine
 */
bool dosleep = false;
/* use '@' and '.' instead of '1' and '0' when dumping the
 * contents of memory
 */
bool dotdash = false;
/* show the contents of memory after each instruction, in emulation mode,
 * of after each tb, in binary translation mode
 */
bool animate = false;
/* during animation, reset the screen before each dump */
bool scrnrst = false;
/* enable binary translation */
bool bintrns = false;
/* enable optimised binary translation */
bool optbntr = false;
/* terminate emulation (set to true if a signal is caught) */
volatile bool stopemul = false;


/* forward declararations */
struct tb_t;
struct cache_t;

/* the mbaby structure contains the virtual machine state that
 * corresponds to the target machine state. For our Manchester
 * Baby Machine this is just the instruction pointer (CI),
 * the accumulator (A) and the memory (M).
 */
struct mbaby {
	static const int MSIZE = 32;
	static const uint32_t ADDR_MASK = MSIZE - 1;
	static const uint32_t OPCODE_MASK = (0x7 << 13);

	int32_t CI;
	int32_t A;
	int32_t M[MSIZE];

	/* constructor: set everything to zero */
	mbaby(): A(0), CI(0) {
		std::fill_n(M, MSIZE, 0);
	}

	/* start emulation. The emulation stops at an halt
	 * instruction (opcode 0x7) or, if bintrns is 1, at the first jump
	 * (opcodes 0x0 and 0x01).  Returns 1 if emulation stopped because
	 * of halt, 0 otherwise.
	 */
	int emulate();
	/* load a program from standard input into memory */
	void load();
	/* dump the contents of memory */
	void dump() const;
	/* start execution using emulation */
	void exec();
	/* start execution using binary translation */
	void btexec();
	/* translate the current dynamic basic block. Returns the
	 * corresponding translated block
	 */
	tb_t* translate() const;

	/* we implement both a 'simple' translation and an
	 * optimized translation. The '-O' option selects the
	 * optimized one.
	 */
	tb_t* translate_simple() const;
	tb_t* translate_opt() const;

	/* number of instructions executed since start */
	unsigned long ninstr;
	/* start of emulation */
	time_point<steady_clock> start;
	/* initialize the start variable */
	void start_time();
	/* possibly slow down the emulation if the
	 * host time is less than the virtual time
	 */
	void slow_down() const;

	tb_t *cur_tb;
};

/* these are the opcodes that the Manchester Baby implements */
int32_t const JMP = 0;		/* JuMp to absolute Pointer  */
int32_t const JRP = 1;		/* Jump to Relative Pointer */
int32_t const LDN = 2;		/* LoaD Negative */
int32_t const STO = 3;		/* STOre */
int32_t const SUB = 4;		/* SUBtract */
int32_t const SUB_alt = 5;	/* alternative opcode for SUBtract */
int32_t const CMP = 6;		/* CoMPare */
int32_t const STP = 7;		/* SToP */

void mbaby::start_time()
{
	ninstr = 0;
	start = steady_clock::now();
}

void mbaby::slow_down() const
{
	if (!dosleep)
		return;

	auto now = steady_clock::now();

	/* number of microseconds elapsed since the beginning of
	 * the emulation (host time)
	 */
	auto ht  = duration_cast<microseconds>(now - start);

	/* number of microseconds elapsed in the target machine
         *
         * The Manchester Baby manipulates one bit a time
         * (it is a serial machine). Reading or writing each
         * bit either from memory, the accumulator or the CI
         * and PI registers takes 10us. Moreover, each 32b word
         * is followed by 4 padding bits. Finally, the execution
         * of each instruction is equivalent to four full reads
         * of 32b word (including the padding bits).
         */
	auto tt = ninstr * (10us * (32 + 4 /* padding */) * 4);

	/* if we are going too fast, we go to sleep */
	if (tt > ht) {
		auto delta = tt - ht;
		if (delta > 1ms)
			usleep(delta.count());
	}
	return;
}

/* for output only: print instruction in symbolic form */
void disas(int32_t instr)
{
	int32_t opcode = (instr & mbaby::OPCODE_MASK) >> 13;
	int32_t addr = instr & mbaby::ADDR_MASK;
	char spaces[] = "                         ";
	std::ostringstream s;

	switch(opcode) {
	case JMP:
		s << "JMP " << addr;
		break;
	case JRP:
		s << "JRP " << addr;
		break;
	case LDN:
		s << "LDN " << addr;
		break;
	case STO:
		s << "STO " << addr;
		break;
	case SUB:
	case SUB_alt:
		s << "SUB " << addr;
		break;
	case CMP:
		s << "CMP";
		break;
	case STP:
		s << "STP";
		break;
	}
	int nsp = sizeof(spaces) - s.str().size() - 1;
	if (nsp < 0)
		nsp = 0;
	spaces[nsp] = '\0';
	std::cout << s.str() << spaces;
}

/* restore the terminal cursor (used when animate is active) */
void stop(int signum)
{
	stopemul = true;
}

/* exec using emulation */
void mbaby::exec()
{
	/* emulate() assumes that CI has already been incremented from
	 * the previous step, so we need to increment it a first time
	 * at the beginning
	 */
	CI = (CI + 1) & ADDR_MASK;
	/* if we are here, bintrns is 0, so emulate() will continue until
	 * if finds an halt instruction
	 */
	emulate();
}

#define MAXHOSTCODE 4096

/* translated block.
 * The translated block is initialized with the address of the first guest
 * instruction. Then, the translation for each guest instruction must
 * be added using addinstr(). Additional code (e.g., prologue and epilogue)
 * may be added with addcode().
 *
 * For simplicity, we pre-allocate a sufficiently large amount of
 * space in the code area when the translation begins, then we
 * release the unused part when the translation ends.
 */
struct tb_t {
	/* the size of the pre-allocated space in the code area */
	static const int TMPCODE = 512;

	/* range of guest addresses of the dynamic basic block */
	int32_t g_beg;
	int32_t g_end;

	/* range of bytes that contain the translated block.
	 * These initially point into the temporary buffer. When
	 * install() is called, they point into the cache
	 */
	uint8_t* h_beg;
	uint8_t* h_end;

	/* start of the host code of the last translated instruction
	 * (used by the patch methods that use relative offsets)
	 */
	uint8_t* h_cur;

	/* init a new tb for guest code that starts at ci */
	tb_t(int32_t ci);

	~tb_t();

	/* return true iff the tb contains the guest address g_a */
	bool contains(int32_t g_a) const { return g_a >= g_beg && g_a < g_end; }

	/* add code (not related to any guest instruction) to the tb */
	void addcode(uint8_t const* code, int size);

	/* convenience helper to avoid passing the size of the code array */
	template <int N> void addcode(uint8_t const (&code)[N]) { addcode(code, N); }

	/* add the translation for another guest instruction */
	void addinstr(uint8_t const* code, int size);

	/* convenience helper to avoid passing the size of the code array */
	template <int N> void addinstr(uint8_t const (&code)[N]) { addinstr(code, N); }

	void call_trampoline(int offset);

	/* return the absolute offset (from the beginning of the tb) of the byte
	 * at given offset in the last addinstr()
	 */
	int absoff(int offset) const;

	/* patch the tb by putting a at given offset */
	template<typename T>
	void abspatch(int offset, T a);

	/* patch the code added by the last addinstr(). put a at given offset */
	template<typename T>
	void patch(int offset, T a);

	/* exec the tb */
	int exec(mbaby* env, uint32_t cache_bitmap)
	{
		/* we cast the h_beg address into a pointer to a function,
		 * then we call the function
		 *
		 * The function receives a pointer to mbaby, which can
		 * be used to reach the emulated registers and memory of the
		 * target machine, and a cache_bitmap, which can be used
		 * to detect self-modifying code
		 */
		env->cur_tb = this;
		if (verbose > 3) {
			std::cout << "exec [" << g_beg << ", " << g_end - 1 << "]\n";
		}
		typedef int (*tb_f)(mbaby* env, uint32_t);
		return (reinterpret_cast<tb_f>(h_beg))(env, cache_bitmap);
	}

	/* show the contents of the TB on stdout */
	void dump() const;
};

void tb_t::addcode(const uint8_t *c, int n)
{
	assert((h_end - h_beg) + n <= TMPCODE);

	/* copy the code into the temporary buffer starting at the end.
	 * Move the h_end pointer to the new end
	 */
	h_end = std::copy(c, c + n, h_end);
}

void tb_t::addinstr(const uint8_t* c, int n)
{
	/* we are adding the translation of a guest instruction.
	 * In addition to copying the code into the temporary buffer
	 * we also need to let h_cur point to the beginning of the
	 * new code, so that patch() below may use it,
	 * and update the range of our dynamic basic block
	 */
	h_cur = h_end;
	addcode(c, n);
	g_end++;
}

int tb_t::absoff(int o) const
{
	return (h_cur - h_beg) + o;
}

template<typename T>
void tb_t::abspatch(int o, T a)
{
	T* p = reinterpret_cast<T*>(h_beg + o);
	*p = a;
}

template<typename T>
void tb_t::patch(int o, T a)
{
	abspatch(absoff(o), a);
}

void tb_t::dump() const
{
	using namespace std;

	ios oldstate(nullptr);
	oldstate.copyfmt(cout);
	cout << "DBB: [" << g_beg << ", " << g_end - 1 << "]\n";
	cout << "TB:\n";
	int col = 0;
	cout << hex << setfill('0');
	for (uint8_t* scan = h_beg; scan != h_end; scan++) {
		if (col == 0)
			cout << "    " << setw(4) << (int)(scan - h_beg) << "  ";
		cout << setw(2) << (int)*scan << " ";
		col++;
		if (col == 16) {
			col = 0;
			cout << "\n";
		}
	}
	if (col != 0)
		cout << "\n";
	csh handle;
	cs_insn* insn;
	if (cs_open(CS_ARCH_X86, CS_MODE_64, &handle) != CS_ERR_OK)
		return;
	cs_option(handle, CS_OPT_SYNTAX, CS_OPT_SYNTAX_ATT);
	size_t count = cs_disasm(handle, h_beg, h_end - h_beg,
		reinterpret_cast<uint64_t>(h_beg), 0, &insn);
	if (count <= 0)
		return;
	cout << "Disassembly:\n";
	for (int i = 0; i < count; i++) {
		cout << "    " << setw(4)
			<< insn[i].address  << "  "
			<< insn[i].mnemonic << ' '
		       	<< insn[i].op_str   << "\n";
	}
	cs_free(insn, count);
	cs_close(&handle);
	cout.copyfmt(oldstate);
}

/* The cache maps Dynamic Basic Blocks (DBB) to Translated Blocks (TB) */
struct cache_t {
	static const int CSIZE = mbaby::MSIZE;
	/* given the start address of a DBB we need to lookup the
	 * corresponding TB in the cache. Since the guest memory is so small,
	 * we can use a very simple data structure: an array with an entry
	 * for each possibile guest address.  If the entry is nullptr, we have no
	 * TB for the given DBB. Otherwise, the entry points to the TB.
	 */
	tb_t* cache[CSIZE];
	/* to detect self-modifying code we need to know all the guest addresses
	 * for which we are storing a translation. Again, since there are only
	 * 32 possible addresses, we can use a simple bitmap with a bit for
	 * each guest address.
	 */
	uint32_t bitmap;
	/* code points to the beginning of the code memory area,
	 * nextcode to the end */
	uint8_t* code;
	uint8_t* nextcode;

	cache_t();

	~cache_t();

	tb_t* find(int32_t CI) const
	{
		return cache[CI];
	}

	/* allocate/free space in the code area */
	uint8_t* allocate(size_t);
	void free(uint8_t*, size_t);

	/* add a new TB to the cache */
	void add(tb_t* tb);
	/* invalidate all the cache */
	void clear();
	/* invalidate only the TBs that contain the given guest addr */
	void invalidate(int32_t addr);

	/* add a "trampoline" code at the beginning of the code area. It is
	 * a function called by the TBs when they have to invalidate the cache
	 */
	void add_trampoline();
	uint8_t* trampoline_entry;
	size_t trampoline_size;

	void dump() const;
};

/* this is the global cache */
cache_t cache;

/* the constructor and destructor of the tb_t must interact
 * with the cache to allocate/free space in the code area
 */
tb_t::tb_t(int32_t ci):
	h_cur(nullptr),
	g_beg(ci),
	g_end(ci)
{
	h_beg = h_end = cache.allocate(TMPCODE);
}

tb_t::~tb_t()
{
	cache.free(h_end, TMPCODE - (h_end - h_beg));
}

void tb_t::call_trampoline(int o)
{
	abspatch(absoff(o), static_cast<uint32_t>(cache.trampoline_entry - (h_beg + absoff(o) + 4)));
}

extern "C" char trampoline, trampoline_end, trampoline_opt, trampoline_opt_end;
cache_t::cache_t(): bitmap(0)
{
	/* we use mmap() to allocate the cache code area, since we need
	 * the execution permission on the corresponding pages
	 */
	code = static_cast<uint8_t*>(mmap(NULL,			// let the system choose the address
				MAXHOSTCODE,			// requested size of the memory area
				PROT_READ|PROT_WRITE|PROT_EXEC, // which permission bits we want
				MAP_SHARED|MAP_ANONYMOUS,	// this area is not backed by a file
				-1,				// no file descriptor
				0				// offset (ignored)
			));
	nextcode = code;

	/* the cache initially contains nothing */
	std::fill_n(cache, CSIZE, nullptr);
}

void cache_t::add_trampoline()
{
	/* copy the trampoline at the beginning */
	char *src = (optbntr ? &trampoline_opt : &trampoline);
	char *end = (optbntr ? &trampoline_opt_end: &trampoline_end);
	trampoline_size = end - src;
	trampoline_entry = nextcode;
	nextcode = std::copy(src, end, trampoline_entry);
}

cache_t::~cache_t()
{
	clear();
	munmap(code, MAXHOSTCODE);
}

uint8_t* cache_t::allocate(size_t n)
{
	assert(n <= MAXHOSTCODE);

	 /* We always add code to the end and, when the cache fills
	  * up, we simply invalidate everthing and start over.
	  */
	if ((nextcode - code) + n > MAXHOSTCODE)
		clear();

	uint8_t *rv = nextcode;
	nextcode += n;
	return rv;
}

void cache_t::free(uint8_t* p, size_t n)
{
	assert(code <= p && p < nextcode);

	if (p + n == nextcode)
		nextcode = p;
}

void cache_t::dump() const
{
	using namespace std;

	ios oldstate(nullptr);
	oldstate.copyfmt(cout);
	cout << "CACHE:\n";
	for (int i = 0; i < CSIZE; i++) {
		if (cache[i]) {
			cout << "    [" << i << "]" << "-> [" << cache[i]->g_beg
				<< ", " << cache[i]->g_end -1 << "]\n";
		}
	}
	cout << "    bitmap " << hex << bitmap << "\n";
	cout.copyfmt(oldstate);
}

void cache_t::add(tb_t* tb)
{
	/* collect the guest locations that are translated by this tb */
	for (int32_t scan = tb->g_beg; scan != tb->g_end; scan++)
		bitmap |= (1U << scan);
	/* install the modified tb in the cache */
	cache[tb->g_beg] = tb;
	if (verbose > 3)
		dump();
}

void cache_t::clear()
{
	for (int i = 0; i < CSIZE; i++) {
		delete cache[i];
		cache[i] = nullptr;
	}
	bitmap = 0;
	nextcode = code + trampoline_size;
}

void cache_t::invalidate(int32_t addr)
{
	/* there may be more than one TB that contains a translation of addr,
	 * so we need to find and invalidate all of them.
	 *
	 * The bitmap is goint to change, so we recompute it along the way.
	 */
	if (verbose > 3)
		std::cout << "invalidate(" << addr << "): ";
	bitmap = 0;
	for (int32_t scan = 0; scan < CSIZE; scan++) {
		tb_t* tb = cache[scan];

		if (!tb)
			continue;

		if (tb->contains(addr)) {
			if (verbose > 3)
				std::cout << "[" << tb->g_beg << ", " << tb->g_end - 1 << "] ";
			cache[scan] = nullptr;
			delete tb;
			continue;
		}

		/* we are keeping this TB, so add it to the bitmap */
		for (int i = tb->g_beg; i < tb->g_end; i++)
			bitmap |= (1U << i);
	}
	if (verbose > 3) {
		std::cout << "\n";
		dump();
	}
}

/* emulate() is used in emulation mode to run until end, but also
 * in binary translation mode to step out of a DBB that modifies itself.
 */
int mbaby::emulate()
{
	while (!stopemul) {

		int32_t PI = M[CI];
		int32_t opcode = (PI & OPCODE_MASK) >> 13;
		int32_t addr = PI & ADDR_MASK;

		if (scrnrst) {
			/* cursor in home position */
			std::cout << "\033[0;0f";
		}
		if (animate) {
			dump();
		}

		switch (opcode) {
		case JMP:
			CI = M[addr];
			/* if bintrns is active we stop at the end of the DBB
			 * (i.e., at the first jump)
			 */
			if (bintrns)
				return 0;
			break;
		case JRP:
			CI = CI + M[addr];
			/* same as above */
			if (bintrns)
				return 0;
			break;
		case LDN:
			A = -M[addr];
			break;
		case STO:
			/* this instruction modifies memory and it may
			 * potentially change a guest instruction. If bintrns
			 * is active we need to take care of the cache
			 */
			if (bintrns && ((1U << addr) & cache.bitmap))
				cache.invalidate(addr);
			M[addr] = A;
			break;
		case SUB:
		case SUB_alt:
			A = A - M[addr];
			break;
		case CMP:
			if (A < 0) {
				CI = CI + 1;
			}
			/* again, stop here if bintrns is active */
			if (bintrns)
				return 0;
			break;
		case STP:
			return 1;
		}

		if (!bintrns) {
			/* a new instrction has been executed */
			ninstr++;

			/* slow down the program if requested */
			slow_down();
		}

		/* the Manchester Baby always increments CI before fetching */
		CI = (CI + 1) & ADDR_MASK;
	}
	return 1;
}


extern "C" bool selfmodify(mbaby *env, int8_t addr)
{
	if (env->cur_tb->contains(addr))
		return true;
	cache.invalidate(addr);
	return false;
}

tb_t* mbaby::translate_simple() const
{
	/* we create a TB that looks like a C++ function, with the proper
	 * prologue and epilogue. The function has the following signature:
	 *
	 * int tb(mbaby *env, uint32_t cache_bitmap);
	 *
	 * The return value has the following interpretation:
	 * 0 		normal end of DBB
	 * 1		STP instruction (end of execution)
	 */

	/* in the prologue we initialize some registers so that
	 * they point to the usefult parts of the virtual machine
	 * state:
	 * %rdi points to CI (equal to env)
	 * %rdx points to A
	 * %rcx points to M
	 * %esi contains the cache_bitmap
	 *
	 * Note that we already receive the correct values for %rdi
	 * and %esi from the caller, so we only need to setup %rdx
	 * and %rcx.
	 *
	 * We also allocate a local variable (in -8(%rbp)) to hold
	 * the return value
	 */
	static const uint8_t prologue[] = {
		0x55,			// pushq %rbp
		0x48, 0x89, 0xe5,	// movq %rsp, %rbp
		0x48, 0x83, 0xec, 0x08,	// subq $8, %rsp
	        0x48, 0x8d, 0x57, 0x04, // leaq 4(%rdi), %rdx
	        0x48, 0x8d, 0x4f, 0x08, // leaq 8(%rdi), %rcx
		0x48, 0xc7, 0x45, 0xf8,
		0x00, 0x00, 0x00, 0x00  // movq $0, -8(%rbp)
	};

	/* in the epilogue we copy the return value in %rax for the caller
	 * and do the standard cleanup
	 */
	static const uint8_t epilogue[] = {
		0x48, 0x8b, 0x45, 0xf8, // movq -8(%rbp), %rax
		0xc9,			// leave
		0xc3			// ret
	};

	/* this is used after the translation of each instruction,
	 * to increment the emulated CI
	 */
	static const uint8_t incCI[] = {
		0xff, 0x07		// incl (%rdi)
	};

	/* ci is the address of the guest instruction we are translating.
	 * At the beginning of the translation it is equal to the current
	 * value of CI, the guest instruction that the virtual machine is
	 * about to execute
	 */
	int32_t ci = CI;
	/* create a new TB for a DBB that starts at ci */
	tb_t* tb = new tb_t(ci);

	/* add the prologue above */
	tb->addcode(prologue);

	/* append a translation for all instructions up to the first jump.  The
	 * general idea is to have a "template" of the host binary code
	 * equivalent to each guest opcode. The template contains some
	 * "variables" that depend on the value of the 'addr' field in the
	 * specific guest instruction that we are translating.  These variables
	 * are implemented as "empty fields" (zero bytes) inside the template.
	 * The template is first added to TB as-is, then the empty fields are
	 * filled by patching the TB.
	 */
	for (;;) {

		int32_t PI = M[ci];
		int32_t opcode = (PI & OPCODE_MASK) >> 13;
		int32_t addr = PI & ADDR_MASK;

		switch (opcode) {
			case JMP: { 	// CI = M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x81,
					0x00, 0x00, 0x00, 0x00,	// movl X(%rcx), %eax
					0x89, 0x07		// movl %eax, (%edi)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				goto out;
			}
			case JRP: {	// CI = CI + M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x81,
					0x00, 0x00, 0x00, 0x00, // movl X(%rcx), %eax
					0x01, 0x07              // addl %eax, (%rdi)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				goto out;
			}
			case LDN: {	// A = -M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x81,
					0x00, 0x00, 0x00, 0x00,	// movl X(%rcx), %eax
					0xf7, 0xd8,             // negl %eax
					0x89, 0x02,            	// movl %eax, (%rdx)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				break;
			}
			case STO: {	// M[addr] = A
				/* this instruction modifies memory, so it may
				 * change an instruction which is already in the cache.
				 * The translated code must first check the cache_bitmap
				 * with the destination address. If there is a match,
				 * the affected TBs must be invalidated before the target
				 * exection may continue.
				 */
				static const uint8_t templ[] = {
					0xf7, 0xc6,
					0x00, 0x00, 0x00, 0x00,	 //    testl $Y, %esi
					0x74, 0x0a,		 //    jz 1f
					0x66, 0x41, 0xb8, 	 
					0x00, 0x00,		 //    movw $Z, %r8w
					0xe8,
					0x00, 0x00, 0x00, 0x00,	 //    call trampoline
					0x8b, 0x02,            	 // 1: movl (%rdx), %eax
					0x89, 0x81,
					0x00, 0x00, 0x00, 0x00   //    movl %eax, X(%rcx)
				};
				tb->addinstr(templ);
				tb->patch(2, 1U << addr);	// set the value of Y
				tb->patch(11, static_cast<int8_t>(addr)); // set the value of Z
				tb->call_trampoline(14);
				tb->patch(22, addr * 4);	// set the value of X
				break;
			}
			case SUB:
			case SUB_alt: {	// A = A - M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x81,
					0x00, 0x00, 0x00, 0x00, // movl X(%rcx), %eax
					0x29, 0x02              // subl %eax, (%rdx)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				break;
			}
			case CMP: {	// if (A < 0) CI = CI + 1
				static const uint8_t templ[] = {
					0x83, 0x3a, 0x00,	//     cmpl $0, (%rdx)
					0x7d, 0x02,   		//     jge 1f
					0xff, 0x07   		//     incl (%rdi)
								// 1:
				};
				tb->addinstr(templ);
				goto out;
			}
			case STP: {	// halt
				static const uint8_t templ[] = {
					0x48, 0xc7, 0x45, 0xf8,
					0x01, 0x00, 0x00, 0x00	// movq $1, -8(%rbp)
				};
				tb->addinstr(templ);
				goto out;
			}
		}

		/* special case: if we reach the end of the memory we stop the DBB */
		if (ci == MSIZE - 1)
			break;

		/* add the code for the increment of CI */
		tb->addcode(incCI);

		/* next instruction to translate */
		ci++;
	}
out:
	tb->addcode(epilogue);

	if (verbose > 3)
		tb->dump();

	return tb;
}

tb_t* mbaby::translate_opt() const
{
	/* Same interface and same structure as translate_simple(), but we use
	 * the register-mapping optimisation: In the prologue we copy the
	 * current values of A and CI into machine registers, so that the
	 * translated instructions can work on them directly; in the epilogue
	 * we copy the registers back into A and CI.
	 *
	 * We also reserve the %rax register to hold the return value, instead
	 * of allocating a local variable.
	 *
	 * Other minor optimisations:
	 * - more compat code (one byte immediates/offsets where possibile)
	 * - cumulative increments to CI
	 *
	 * By looking at output DBBs we can easily see that there is still
	 * a lot of room for further optimisations, but these would require
	 * an analysis of the TB as a whole.
	 */

	/*
	 * %rdi points to M
	 * %esi contains the cache bitmap
	 * %ecx is a local copy of CI
	 * %edx ia a local copy of A
	 *
	 */
	static const uint8_t prologue[] = {
		0x8b, 0x0f,		// movl (%rdi), %ecx
		0x8b, 0x57, 0x04,	// movl 4(%rdi), %edx
		0x48, 0x83, 0xc7, 0x08, // addq $8, %rdi
		0x48, 0x31, 0xc0	// xorq %rax, %rax
	};

	/* int the epilogue we copy %ecx and %edx back into
	 * the CI and A fields of the env.
	 * Note that %rax already contains the return value.
	 */
	static const uint8_t epilogue[] = {
		0x48, 0x83, 0xef, 0x08, // subq $8, %rdi
		0x89, 0x0f,		// movl %ecx, (%rdi)
		0x89, 0x57, 0x04,	// movl %edx, 4(%rdi)
		0xc3			// ret
	};

	/* this will be placed at the end of the DBB, before the epilogue,
	 * to increment the local copy of CI by the total number emulated
	 * guest instructions
	 */
	static const uint8_t addCI[] = {
		0x83, 0xc1, 0x00		// addl $X, %ecx
	};

	/* as a minor optimisation, we try to avoid reloading A from
	 * memory if we are loading from the same address of the last
	 * store, and A has not changed in the meantime.
	 *
	 * We remember the address of the last store and compare it
	 * with the address of the load. We also mark the address as
	 * invalid whenever A is written.
	 */
	int32_t last_store_addr;	// -1 if not valid

	int32_t ci = CI;
	tb_t* tb = new tb_t(ci);

	tb->addcode(prologue);

	for (;;) {

		int32_t PI = M[ci];
		int32_t opcode = (PI & OPCODE_MASK) >> 13;
		int32_t addr = PI & ADDR_MASK;

		switch (opcode) {
			case JMP: { 	// CI = M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x4f, 0x00, 	//    movl X(%rdi), %ecx
				};
				tb->addinstr(templ);
				tb->patch(2, static_cast<int8_t>(addr * 4));	// set the value of X
				/* skip the addCI, since this instructions puts an
				 * absolute value in CI
				 */
				goto out2;
			}
			case JRP: {	// CI = CI + M[addr]
				static const uint8_t templ[] = {
					0x03, 0x4f, 0x00, 	//    addl X(%rdi), %ecx
				};
				tb->addinstr(templ);
				tb->patch(2, static_cast<int8_t>(addr * 4)); // set the value of X
				goto out;
			}
			case LDN: {	// A = -M[addr]
				if (last_store_addr != addr) {
					/* normal case */
					static const uint8_t templ[] = {
						0x8b, 0x57, 0x00, 	//    movl X(%rdi), %edx
						0xf7, 0xda,             //    negl %edx
					};
					tb->addinstr(templ);
					tb->patch(2, static_cast<int8_t>(addr * 4)); // set the value of X
				} else {
					/* optimised case: %edx already contains the
					 * value to be loaded from memory, so we
					 * skip the load.
					 */
					static const uint8_t templ[] = {
						0xf7, 0xda,		// negl %edx
					};
					tb->addinstr(templ);
				}
				/* A is written, invalidate the last_store_addr */
				last_store_addr = -1;
				break;
			}
			case STO: {	// M[addr] = A
				static const uint8_t templ[] = {
					0x0f, 0xba, 0xe6, 0x00, //    btl $X, %esi
					0x73, 0xa,		//    jnc 1f
					0x66, 0x41, 0xb8,	//    movw $Z, %r8w
					0x00, 0x00,
					0xe8,
					0x00, 0x00, 0x00, 0x00, //    call trampoline
					0x89, 0x57, 0x00, 	// 1: movl %edx, K(%rdi)
				};
				tb->addinstr(templ);
				tb->patch(3, static_cast<int8_t>(addr)); // set the value of X
				tb->patch(9, static_cast<uint16_t>(addr | ((ci - CI) << 8U)));
				tb->call_trampoline(12);
				tb->patch(18, static_cast<int8_t>(addr * 4));	// set the value of Z
				/* remember the last_stor_addr */
				last_store_addr = addr;
				break;
			}
			case SUB:
			case SUB_alt: {	// A = A - M[addr]
				static const uint8_t templ[] = {
					0x2b, 0x57, 0x00	//    subl X(%rdi), %edx
				};
				tb->addinstr(templ);
				tb->patch(2, static_cast<int8_t>(addr * 4));	// set the value of X
				/* A is written, invalidate the last_store_addr */
				last_store_addr = -1;
				break;
			}
			case CMP: {	// if (A < 0) CI = CI + 1
				static const uint8_t templ[] = {
					0x83, 0xfa, 0x00,	//    cmpl $0, %edx
					0x7d, 0x2,		//    jge 1f
					0xff, 0xc1		//    incl %ecx
								// 1:
				};
				tb->addinstr(templ);
				goto out;
			}
			case STP: {	// halt
				static const uint8_t templ[] = {
					0xb0, 0x01		//    movb $1, %al
				};
				tb->addinstr(templ);
				goto out;
			}
		}

		/* special case: if we reach the end of the memory we stop the DBB */
		if (ci == MSIZE - 1)
			break;

		/* next instruction to translate */
		ci++;
	}
out:
	if (ci - CI) { // avoid the output of 'add $0, %rcx'
		int aoff = (tb->h_end - tb->h_beg);
		tb->addcode(addCI);
		tb->abspatch(aoff + 2, static_cast<int8_t>(ci - CI));
	}
out2:
	tb->addcode(epilogue);

	if (verbose > 3)
		tb->dump();

	return tb;
}

tb_t* mbaby::translate() const
{
	return optbntr ? translate_opt(): translate_simple();
}

void mbaby::btexec()
{
	tb_t* tb;

	while (!stopemul) {
		CI = (CI + 1) & ADDR_MASK;

		if (scrnrst) {
			/* cursor in home position */
			std::cout << "\033[0;0f";
		}

		if (animate)
			dump();

		tb = cache.find(CI);
		if (!tb) {
			tb = translate();
			cache.add(tb);
		}

		unsigned int tbsz = tb->g_end - tb->g_beg;

		bool stop = tb->exec(this, cache.bitmap);

		ninstr += tbsz;
		slow_down();

		if (stop)
			return;
	}
}

struct bin2ascii {
	int32_t v_;
	bin2ascii(int32_t v): v_(v) {};
};

std::ostream& operator <<(std::ostream& o, const bin2ascii& b)
{
	uint32_t v_ = static_cast<uint32_t>(b.v_);

	for (int i = 0; i < 32; i++) {
		o << static_cast<char>(dotdash ? ((v_ & 0x1) ? '@' : '.') :
						  ('0' + (v_ & 0x1)));
		v_ >>= 0x1;
	}
	return o;
}

int32_t ascii2bin(std::string::iterator beg, std::string::iterator end)
{
	uint32_t v = 0;

	for (int i = 0; i < 32; i++) {
		v >>= 0x1;
		if (beg != end && *beg++ == '1')
			v |= 0x80000000;
	}
	return (int32_t)v;
}

void mbaby::dump() const
{
	using namespace std;

	for (int i = 0; i < MSIZE; i++) {
		if (verbose)
			cout << setw(4) << i << "  ";
		cout << bin2ascii(M[i]);
		if (verbose > 1) {
			cout << " // " << setw(8) << setbase(16) << M[i] << " " <<
			     setw(16) << setbase(10) << M[i] << " " << (CI == i ? '*' : ' ');
			disas(M[i]);
		}
		cout << "\n";
	}
	if (verbose > 2) {
		cout << '\n';
		cout << "  A   " << bin2ascii(A) <<
			" // " << setw(8) << setbase(16) << A << " " <<
		     setw(16) << setbase(10) << A << "  ";
		disas(A);
	        cout << '\n';
		cout << "  CI  " << bin2ascii(CI) <<
			" // " << setw(8) << setbase(16) << CI << " " <<
		     setw(16) << setbase(10) << CI << '\n';
		cout << "\n";
	}
	if (scrnrst)
		cout.flush();
}

void mbaby::load()
{
	std::string line;

	for (int i = 0; i < MSIZE; i++) {
		M[i] = getline(std::cin, line) ? ascii2bin(line.begin(), line.end()) : 0;
	}
}

void usage(const char* name)
{
	std::cerr << "Usage: " << name << " [-advnsrbO]\n";
}

int
main(int argc, char* argv[])
{
	char ch;
	bool noexec = false;
	mbaby m;


	while ( (ch = getopt(argc, argv, "advnsrbO")) != -1) {
		switch (ch) {
		default:
			usage(argv[0]);
			return 1;
		case 'd':
			dotdash = true;
			break;
		case 'v':
			verbose++;
			break;
		case 'n':
			noexec = true;
			break;
		case 's':
			dosleep = true;
			break;
		case 'a':
			animate = true;
			break;
		case 'r':
			scrnrst = true;
			animate = true;
			break;
		case 'b':
			bintrns = true;
			break;
		case 'O':
			bintrns = true;
			optbntr = true;
			break;
		}
	}
	if (scrnrst)
		std::ios::sync_with_stdio(false);

	if (bintrns) {
		cache.add_trampoline();
	}
	m.load();
	if (verbose) {
		if (!noexec)
			std::cout << "Initial contents of memory:\n";
		m.dump();
	}
	if (noexec)
		return 0;
	if (scrnrst) {
		/* clear the screen */
		std::cout << "\033[2J";
		/* hide cursor */
		std::cout << "\033[?25l";
	}
	/* intercept common signal and stop emulation */
	signal(SIGTERM, stop);
	signal(SIGINT, stop);
	signal(SIGQUIT, stop);

	m.start_time();

	if (bintrns)
		m.btexec();
	else
		m.exec();
	if (verbose && !scrnrst)
		std::cout << "Final contents of memory:\n";
	if (scrnrst) {
		/* cursor in home position */
		std::cout << "\033[0;0f";
	}
	m.dump();
	if (scrnrst) {
		/* show the cursor */
		std::cout << "\033[?25h";
	}
	return 0;
}
